import datetime
import uuid

import pytz
from django.core.management import call_command
from django.test import TestCase, override_settings
from freezegun import freeze_time

from silk import models
from silk.config import SilkyConfig
from silk.storage import ProfilerResultStorage

from .factories import RequestMinFactory, ResponseFactory, SQLQueryFactory

# TODO test atomicity

# http://stackoverflow.com/questions/13397038/uuid-max-character-length
# UUID_MAX_LENGTH = 36

# TODO move to separate file test and collection it self


class CaseInsensitiveDictionaryTest:
    pass


class RequestTest(TestCase):

    def setUp(self):

        self.obj = RequestMinFactory.create()
        self.max_percent = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT
        self.max_requests = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS

    def tearDown(self):

        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = self.max_percent
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = self.max_requests

    def test_uuid_is_primary_key(self):

        self.assertIsInstance(self.obj.id, uuid.UUID)

    @freeze_time('2016-01-01 12:00:00')
    def test_start_time_field_default(self):

        obj = RequestMinFactory.create()
        self.assertEqual(obj.start_time, datetime.datetime(2016, 1, 1, 12, 0, 0, tzinfo=pytz.UTC))

    def test_total_meta_time_if_have_no_meta_and_queries_time(self):

        self.assertEqual(self.obj.total_meta_time, 0)

    def test_total_meta_time_if_have_meta_time_spent_queries(self):

        obj = RequestMinFactory.create(meta_time_spent_queries=10.5)
        self.assertEqual(obj.total_meta_time, 10.5)

    def test_total_meta_time_if_meta_time(self):

        obj = RequestMinFactory.create(meta_time=3.3)
        self.assertEqual(obj.total_meta_time, 3.3)

    def test_total_meta_if_self_have_it_meta_and_queries_time(self):

        obj = RequestMinFactory.create(meta_time=3.3, meta_time_spent_queries=10.5)
        self.assertEqual(obj.total_meta_time, 13.8)

    def test_time_spent_on_sql_queries_if_has_no_related_SQLQueries(self):

        self.assertEqual(self.obj.time_spent_on_sql_queries, 0)

    # FIXME probably a bug
    def test_time_spent_on_sql_queries_if_has_related_SQLQueries_with_no_time_taken(self):

        query = SQLQueryFactory()
        self.obj.queries.add(query)

        self.assertEqual(query.time_taken, None)

        with self.assertRaises(TypeError):
            self.obj.time_spent_on_sql_queries

    def test_time_spent_on_sql_queries_if_has_related_SQLQueries_and_time_taken(self):

        query1 = SQLQueryFactory(time_taken=3.5)
        query2 = SQLQueryFactory(time_taken=1.5)
        RequestMinFactory().queries.add(query1, query2)

        self.assertEqual(self.obj.time_spent_on_sql_queries, 0)

    def test_headers_if_has_no_encoded_headers(self):

        self.assertIsInstance(self.obj.headers, models.CaseInsensitiveDictionary)
        self.assertFalse(self.obj.headers)

    def test_headers_if_has_encoded_headers(self):

        self.obj.encoded_headers = '{"some-header": "some_data"}'
        self.assertIsInstance(self.obj.headers, models.CaseInsensitiveDictionary)
        self.assertDictEqual(self.obj.headers, {'some-header': 'some_data'})

    def test_content_type_if_no_headers(self):

        self.assertEqual(self.obj.content_type, None)

    def test_content_type_if_no_specific_content_type(self):

        self.obj.encoded_headers = '{"foo": "some_data"}'
        self.assertEqual(self.obj.content_type, None)

    def test_content_type_if_header_have_content_type(self):

        self.obj.encoded_headers = '{"content-type": "some_data"}'
        self.assertEqual(self.obj.content_type, "some_data")

    def test_garbage_collect(self):

        self.assertTrue(models.Request.objects.filter(id=self.obj.id).exists())
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = 100
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = 0
        models.Request.garbage_collect()
        self.assertFalse(models.Request.objects.filter(id=self.obj.id).exists())

    def test_probabilistic_garbage_collect(self):

        self.assertTrue(models.Request.objects.filter(id=self.obj.id).exists())
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = 0
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = 0
        models.Request.garbage_collect()
        self.assertTrue(models.Request.objects.filter(id=self.obj.id).exists())

    def test_force_garbage_collect(self):

        self.assertTrue(models.Request.objects.filter(id=self.obj.id).exists())
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = 0
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = 0
        models.Request.garbage_collect(force=True)
        self.assertFalse(models.Request.objects.filter(id=self.obj.id).exists())

    def test_greedy_garbage_collect(self):

        for x in range(3):
            obj = models.Request(path='/', method='get')
            obj.save()
        self.assertEqual(models.Request.objects.count(), 4)
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = 50
        SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = 3
        models.Request.garbage_collect(force=True)
        self.assertGreater(models.Request.objects.count(), 0)

    def test_save_if_have_no_raw_body(self):

        obj = models.Request(path='/some/path/', method='get')
        self.assertEqual(obj.raw_body, '')
        obj.save()
        self.assertEqual(obj.raw_body, '')

    def test_save_if_have_raw_body(self):

        obj = models.Request(path='/some/path/', method='get', raw_body='some text')
        obj.save()
        self.assertEqual(obj.raw_body, 'some text')

    def test_save_if_have_no_body(self):

        obj = models.Request(path='/some/path/', method='get')
        self.assertEqual(obj.body, '')
        obj.save()
        self.assertEqual(obj.body, '')

    def test_save_if_have_body(self):

        obj = models.Request(path='/some/path/', method='get', body='some text')
        obj.save()
        self.assertEqual(obj.body, 'some text')

    def test_save_if_have_no_end_time(self):

        obj = models.Request(path='/some/path/', method='get')
        self.assertEqual(obj.time_taken, None)
        obj.save()
        self.assertEqual(obj.time_taken, None)

    @freeze_time('2016-01-01 12:00:00')
    def test_save_if_have_end_time(self):

        date = datetime.datetime(2016, 1, 1, 12, 0, 3, tzinfo=pytz.UTC)
        obj = models.Request(path='/some/path/', method='get', end_time=date)
        obj.save()
        self.assertEqual(obj.end_time, date)
        self.assertEqual(obj.time_taken, 3000.0)

    def test_prof_file_default_storage(self):
        obj = models.Request(path='/some/path/', method='get')
        self.assertEqual(obj.prof_file.storage.__class__, ProfilerResultStorage)


class ResponseTest(TestCase):

    def setUp(self):

        self.obj = ResponseFactory.create()

    def test_uuid_is_primary_key(self):

        self.assertIsInstance(self.obj.id, uuid.UUID)

    def test_is_1to1_related_to_request(self):

        request = RequestMinFactory.create()
        resp = models.Response.objects.create(status_code=200, request=request)

        self.assertEqual(request.response, resp)

    def test_headers_if_has_no_encoded_headers(self):

        self.assertIsInstance(self.obj.headers, models.CaseInsensitiveDictionary)
        self.assertFalse(self.obj.headers)

    def test_headers_if_has_encoded_headers(self):

        self.obj.encoded_headers = '{"some-header": "some_data"}'
        self.assertIsInstance(self.obj.headers, models.CaseInsensitiveDictionary)
        self.assertDictEqual(self.obj.headers, {'some-header': 'some_data'})

    def test_content_type_if_no_headers(self):

        self.assertEqual(self.obj.content_type, None)

    def test_content_type_if_no_specific_content_type(self):

        self.obj.encoded_headers = '{"foo": "some_data"}'
        self.assertEqual(self.obj.content_type, None)

    def test_content_type_if_header_have_content_type(self):

        self.obj.encoded_headers = '{"content-type": "some_data"}'
        self.assertEqual(self.obj.content_type, "some_data")


class SQLQueryManagerTest(TestCase):

    def test_if_no_args_passed(self):
        pass

    def test_if_one_arg_passed(self):
        pass

    def if_a_few_args_passed(self):
        pass

    def if_objs_kw_arg_passed(self):
        pass

    def if_not_the_objs_kw_arg_passed(self):
        pass


class SQLQueryTest(TestCase):

    def setUp(self):

        self.obj = SQLQueryFactory.create()
        self.end_time = datetime.datetime(2016, 1, 1, 12, 0, 5, tzinfo=pytz.UTC)
        self.start_time = datetime.datetime(2016, 1, 1, 12, 0, 0, tzinfo=pytz.UTC)

    @freeze_time('2016-01-01 12:00:00')
    def test_start_time_field_default(self):

        obj = SQLQueryFactory.create()
        self.assertEqual(obj.start_time, datetime.datetime(2016, 1, 1, 12, 0, 0, tzinfo=pytz.UTC))

    def test_is_m2o_related_to_request(self):

        request = RequestMinFactory()
        self.obj.request = request
        self.obj.save()

        self.assertIn(self.obj, request.queries.all())

    def test_query_manager_instance(self):

        self.assertIsInstance(models.SQLQuery.objects, models.SQLQueryManager)

    def test_traceback_ln_only(self):

        self.obj.traceback = """Traceback (most recent call last):
          File "/home/user/some_script.py", line 10, in some_func
            pass
          File "/usr/lib/python2.7/bdb.py", line 20, in trace_dispatch
            return self.dispatch_return(frame, arg)
          File "/usr/lib/python2.7/bdb.py", line 30, in dispatch_return
            if self.quitting: raise BdbQuit
        BdbQuit"""

        output = ('Traceback (most recent call last):\n'
                  '            pass\n'
                  '            return self.dispatch_return(frame, arg)\n'
                  '            if self.quitting: raise BdbQuit')

        self.assertEqual(self.obj.traceback_ln_only, output)

    def test_formatted_query_if_no_query(self):

        self.obj.query = ""
        self.obj.formatted_query

    def test_formatted_query_if_has_a_query(self):

        query = """SELECT Book.title AS Title,
             COUNT(*) AS Authors
             FROM  Book
             JOIN  Book_author
                 ON  Book.isbn = Book_author.isbn
             GROUP BY Book.title;"""

        self.obj.query = query
        self.obj.formatted_query

    def test_num_joins_if_no_joins_in_query(self):

        query = """SELECT Book.title AS Title,
             COUNT(*) AS Authors
             FROM  Book
             GROUP BY Book.title;"""

        self.obj.query = query

        self.assertEqual(self.obj.num_joins, 0)

    def test_num_joins_if_joins_in_query(self):

        query = """SELECT p.id
                   FROM Person p
                       JOIN address a ON p.Id = a.Person_ID
                       JOIN address_type at ON a.Type_ID = at.Id
                       JOIN `option` o ON p.Id = o.person_Id
                       JOIN option_address_type oat ON o.id = oat.option_id
                    WHERE a.country_id = 1 AND at.id <> oat.type_id;"""

        self.obj.query = query
        self.assertEqual(self.obj.num_joins, 4)

    def test_num_joins_if_no_joins_in_query_but_this_word_searched(self):

        query = """SELECT Book.title FROM  Book WHERE Book.title=`Join the dark side, Luke!`;"""

        self.obj.query = query
        self.assertEqual(self.obj.num_joins, 0)

    def test_num_joins_if_multiple_statement_in_query(self):
        query = """SELECT Book.title FROM  Book WHERE Book.title=`Join the dark side, Luke!`;
                   SELECT Book.joiner FROM Book
                    LEFT OUTER JOIN joined ON Book.joiner = joined.joiner
                    INNER JOIN joined ON Book.joiner = joined.joiner
                   WHERE Book.joiner='Join i am'"""

        self.obj.query = query
        self.assertEqual(self.obj.num_joins, 2)

    def test_tables_involved_if_no_query(self):

        self.obj.query = ''

        self.assertEqual(self.obj.tables_involved, [])

    def test_tables_involved_if_query_has_only_a_from_token(self):

        query = """SELECT * FROM  Book;"""
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Book;'])

    def test_tables_involved_if_query_has_a_join_token(self):

        query = """SELECT p.id FROM Person p JOIN Address a ON p.Id = a.Person_ID;"""
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Person', 'Address'])

    def test_tables_involved_if_query_has_an_as_token(self):

        query = 'SELECT Book.title AS Title FROM  Book GROUP BY Book.title;'
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Title', 'Book'])

    # FIXME bug, not a feature
    def test_tables_involved_check_with_fake_a_from_token(self):

        query = """SELECT * FROM  Book WHERE Book.title=`EVIL FROM WITHIN`;"""
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Book', 'WITHIN`;'])

    # FIXME bug, not a feature
    def test_tables_involved_check_with_fake_a_join_token(self):

        query = """SELECT * FROM  Book WHERE Book.title=`Luke, join the dark side!`;"""
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Book', 'the'])

    # FIXME bug, not a feature
    def test_tables_involved_check_with_fake_an_as_token(self):

        query = """SELECT * FROM  Book WHERE Book.title=`AS SOON AS POSIABLE`;"""
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['Book', 'POSIABLE`;'])

    def test_tables_involved_if_query_has_subquery(self):

        query = '''SELECT A.Col1, A.Col2, B.Col1,B.Col2
                  FROM (SELECT RealTableZ.Col1, RealTableY.Col2, RealTableY.ID AS ID
                          FROM RealTableZ
                               LEFT OUTER JOIN RealTableY ON RealTableZ.ForeignKeyY=RealTableY.ID
                         WHERE RealTableY.Col11>14
                        ) AS B INNER JOIN A
                ON A.ForeignKeyY=B.ID;'''
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['ID', 'RealTableZ', 'RealTableY', 'B', 'A'])

    # FIXME bug, not a feature
    def test_tables_involved_if_query_has_django_aliase_on_column_names(self):

        query = 'SELECT foo AS bar FROM some_table;'
        self.obj.query = query
        self.assertEqual(self.obj.tables_involved, ['bar', 'some_table;'])

    def test_save_if_no_end_and_start_time(self):

        obj = SQLQueryFactory.create()

        self.assertEqual(obj.time_taken, None)

    @freeze_time('2016-01-01 12:00:00')
    def test_save_if_has_end_time(self):

        # datetime.datetime(2016, 1, 1, 12, 0, 5, tzinfo=pytz.UTC)
        obj = SQLQueryFactory.create(end_time=self.end_time)

        self.assertEqual(obj.time_taken, 5000.0)

    @freeze_time('2016-01-01 12:00:00')
    def test_save_if_has_start_time(self):

        obj = SQLQueryFactory.create(start_time=self.start_time)

        self.assertEqual(obj.time_taken, None)

    def test_save_if_has_end_and_start_time(self):

        obj = SQLQueryFactory.create(start_time=self.start_time, end_time=self.end_time)

        self.assertEqual(obj.time_taken, 5000.0)

    def test_save_if_has_pk_and_request(self):

        self.obj.request = RequestMinFactory.create()
        self.obj.save()
        self.assertEqual(self.obj.request.num_sql_queries, 0)

    def test_save_if_has_no_pk(self):

        obj = SQLQueryFactory.build(start_time=self.start_time, end_time=self.end_time)
        obj.request = RequestMinFactory.create()
        obj.save()
        self.assertEqual(obj.request.num_sql_queries, 1)

    # should not rise
    def test_save_if_has_no_request(self):

        obj = SQLQueryFactory.build(start_time=self.start_time, end_time=self.end_time)
        obj.save()

    # FIXME a bug
    def test_delete_if_no_related_requests(self):

        with self.assertRaises(AttributeError):
            self.obj.delete()

        # self.assertNotIn(self.obj, models.SQLQuery.objects.all())

    def test_delete_if_has_request(self):

        self.obj.request = RequestMinFactory.create()
        self.obj.save()
        self.obj.delete()

        self.assertNotIn(self.obj, models.SQLQuery.objects.all())


class NoPendingMigrationsTest(TestCase):
    """
    Test if proper migrations are added and the models state is consistent.
    It should make sure that no new migrations are created for this app,
    when end-user runs `makemigrations` command.
    """

    def test_no_pending_migrations(self):
        call_command("makemigrations", "silk", "--check", "--dry-run")

    @override_settings(DEFAULT_AUTO_FIELD='django.db.models.BigAutoField')
    def test_check_with_overridden_default_auto_field(self):
        """
        Test with `BigAutoField` set as `DEFAULT_AUTO_FIELD` - which is
        default when generating proj with Django 3.2.
        """
        self.test_no_pending_migrations()


class BaseProfileTest(TestCase):
    pass


class ProfileTest(TestCase):
    pass



"""
Test profiling of DB queries without mocking, to catch possible
incompatibility
"""
from django.shortcuts import reverse
from django.test import Client, TestCase

from silk.collector import DataCollector
from silk.config import SilkyConfig
from silk.models import Request
from silk.profiling.profiler import silk_profile

from .factories import BlindFactory


class TestDbQueries(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        BlindFactory.create_batch(size=5)
        SilkyConfig().SILKY_META = False

    def test_profile_request_to_db(self):
        client = Client()
        DataCollector().configure(Request(reverse('example_app:index')))

        with silk_profile(name='test_profile'):
            resp = client.get(reverse('example_app:index'))

        DataCollector().profiles.values()
        assert len(resp.context['blinds']) == 5


class TestAnalyzeQueries(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        BlindFactory.create_batch(size=5)
        SilkyConfig().SILKY_META = False
        SilkyConfig().SILKY_ANALYZE_QUERIES = True

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        SilkyConfig().SILKLY_ANALYZE_QUERIES = False

    def test_analyze_queries(self):
        DataCollector().configure(Request(reverse('example_app:index')))
        client = Client()

        with silk_profile(name='test_profile'):
            resp = client.get(reverse('example_app:index'))

        DataCollector().profiles.values()
        assert len(resp.context['blinds']) == 5


class TestAnalyzeQueriesExplainParams(TestAnalyzeQueries):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        SilkyConfig().SILKY_EXPLAIN_FLAGS = {'verbose': True}

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        SilkyConfig().SILKY_EXPLAIN_FLAGS = None



from time import sleep

# Create your views here.
from django.shortcuts import render
from example_app import models

from silk.profiling.profiler import silk_profile


def index(request):
    @silk_profile()
    def do_something_long():
        sleep(1.345)

    with silk_profile(name='Why do this take so long?'):
        do_something_long()
    return render(request, 'example_app/index.html', {'blinds': models.Blind.objects.all()})



import random

from django.db.models import Count, F
from django.test import TestCase
from django.urls import reverse

from silk import models
from silk.config import SilkyConfig
from silk.middleware import silky_reverse

from .test_lib.mock_suite import MockSuite


class TestEndPoints(TestCase):
    """
    Hit all the endpoints to check everything actually renders/no error 500s etc.
    Each test will ensure that an object with something to display is chosen to be rendered e.g.
    a request/profile that has queries
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # We're not testing auth here.
        SilkyConfig().SILKY_AUTHORISATION = False
        SilkyConfig().SILKY_AUTHENTICATION = False
        mock_suite = MockSuite()
        for _ in range(0, 100):
            mock_suite.mock_request()

    def test_summary(self):
        response = self.client.get(silky_reverse('summary'))
        self.assertTrue(response.status_code == 200)

    def test_requests(self):
        response = self.client.get(silky_reverse('requests'))
        self.assertTrue(response.status_code == 200)

    def test_request_detail(self):
        request_query_data = random.choice(models.Request.objects.values('id'))
        request_id = request_query_data['id']
        response = self.client.get(silky_reverse('request_detail', kwargs={
            'request_id': request_id
        }))
        self.assertTrue(response.status_code == 200)

    def test_request_sql(self):
        request_query_data = random.choice(
            models.SQLQuery.objects
            .values('request_id')
            .filter(request_id__isnull=False)
        )
        request_id = request_query_data['request_id']
        response = self.client.get(silky_reverse('request_sql', kwargs={'request_id': request_id}))
        self.assertTrue(response.status_code == 200)

    def test_request_sql_detail(self):
        kwargs = random.choice(
            models.SQLQuery.objects
            .annotate(sql_id=F('id'))
            .values('sql_id', 'request_id')
            .filter(request_id__isnull=False)
        )
        response = self.client.get(silky_reverse('request_sql_detail', kwargs=kwargs))
        self.assertTrue(response.status_code == 200)

    def test_raw(self):
        request_query_data = random.choice(
            models.Request.objects
            .values('id')
            .filter(body__isnull=False)
        )
        request_id = request_query_data['id']
        url = reverse('silk:raw', kwargs={
            'request_id': request_id
        }) + '?typ=request&subtyp=processed'
        response = self.client.get(url)
        code = response.status_code
        self.assertTrue(code == 200)

    def test_request_profiling(self):
        request_id = random.choice(
            models.Profile.objects
            .values('request_id')
            .filter(request_id__isnull=False)
        )
        response = self.client.get(silky_reverse('request_profiling', kwargs=request_id))
        self.assertTrue(response.status_code == 200)

    def test_request_profile_detail(self):
        kwargs = random.choice(
            models.Profile.objects
            .annotate(profile_id=F('id'))
            .values('profile_id', 'request_id')
            .filter(request_id__isnull=False)
        )
        response = self.client.get(silky_reverse('request_profile_detail', kwargs=kwargs))
        self.assertTrue(response.status_code == 200)

    def test_request_and_profile_sql(self):
        kwargs = random.choice(
            models.Profile.objects
            .annotate(num=Count('queries'), profile_id=F('id'))
            .values('profile_id', 'request_id')
            .filter(request_id__isnull=False, num__gt=0)
        )
        response = self.client.get(silky_reverse('request_and_profile_sql', kwargs=kwargs))
        self.assertTrue(response.status_code == 200)

    def test_request_and_profile_sql_detail(self):
        random_profile = random.choice(
            models.Profile.objects
            .annotate(num=Count('queries'), profile_id=F('id'))
            .values('profile_id', 'request_id')
            .filter(request_id__isnull=False, num__gt=0)
        )
        random_sql_query = random.choice(
            models.SQLQuery.objects
            .annotate(sql_id=F('id'))
            .values('sql_id')
            .filter(profiles__id=random_profile['profile_id'])
        )
        kwargs = {}
        kwargs.update(random_profile)
        kwargs.update(random_sql_query)
        response = self.client.get(silky_reverse('request_and_profile_sql_detail', kwargs=kwargs))
        self.assertTrue(response.status_code == 200)

    def test_profile_detail(self):
        profile_query_data = random.choice(models.Profile.objects.values('id'))
        profile_id = profile_query_data['id']
        response = self.client.get(silky_reverse('profile_detail', kwargs={
            'profile_id': profile_id
        }))
        self.assertTrue(response.status_code == 200)

    def test_profile_sql(self):
        profile_query_data = random.choice(
            models.Profile.objects
            .annotate(num=Count('queries'))
            .values('id')
            .filter(num__gt=0)
        )
        profile_id = profile_query_data['id']
        response = self.client.get(silky_reverse('profile_sql', kwargs={'profile_id': profile_id}))
        self.assertTrue(response.status_code == 200)

    def test_profile_sql_detail(self):
        profile_query_data = random.choice(
            models.Profile.objects
            .annotate(num=Count('queries'))
            .values('id')
            .filter(num__gt=0)
        )
        profile_id = profile_query_data['id']
        sql_id = random.choice(models.SQLQuery.objects.filter(profiles=profile_id)).pk
        response = self.client.get(silky_reverse('profile_sql_detail', kwargs={'profile_id': profile_id,
                                                                               'sql_id': sql_id}))
        self.assertTrue(response.status_code == 200)

    def test_profiling(self):
        response = self.client.get(silky_reverse('profiling'))
        self.assertTrue(response.status_code == 200)



import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = 'ey5!m&h-uj6c7dzp@(o1%96okkq4!&bjja%oi*v3r=2t(!$7os'

DEBUG = True
DEBUG_PROPAGATE_EXCEPTIONS = True

ALLOWED_HOSTS = []

INSTALLED_APPS = (
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'silk',
    'example_app'
)

ROOT_URLCONF = 'project.urls'

# Django 3.2+
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'silk.middleware.SilkyMiddleware'
]

WSGI_APPLICATION = 'wsgi.application'

DB_ENGINE = os.environ.get("DB_ENGINE", "postgresql")

DATABASES = {
    "default": {
        "ENGINE": f"django.db.backends.{DB_ENGINE}",
        "NAME": os.environ.get("DB_NAME", "postgres"),
        "USER": os.environ.get("DB_USER", 'postgres'),
        "PASSWORD": os.environ.get("DB_PASSWORD", "postgres"),
        "HOST": os.environ.get("DB_HOST", "127.0.0.1"),
        "PORT": os.environ.get("DB_PORT", 5432),
        "ATOMIC_REQUESTS": True
    },
}

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True

LOGGING = {
    'version': 1,
    'formatters': {
        'mosayc': {
            'format': '%(asctime)-15s %(levelname)-7s %(message)s [%(funcName)s (%(filename)s:%(lineno)s)]',
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'mosayc'
        }
    },
    'loggers': {
        'silk': {
            'handlers': ['console'],
            'level': 'DEBUG'
        }
    },
}

STATIC_URL = '/static/'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

TEMP_DIR = os.path.join(BASE_DIR, "tmp")
STATIC_ROOT = os.path.join(TEMP_DIR, "static")

if not os.path.exists(STATIC_ROOT):
    os.makedirs(STATIC_ROOT)

MEDIA_ROOT = BASE_DIR + '/media/'
MEDIA_URL = '/media/'

if not os.path.exists(MEDIA_ROOT):
    os.mkdir(MEDIA_ROOT)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

LOGIN_URL = '/login/'
LOGIN_REDIRECT_URL = '/'

SILKY_META = True
SILKY_PYTHON_PROFILER = True
SILKY_PYTHON_PROFILER_BINARY = True
# Do not garbage collect for tests
SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT = 0
# SILKY_AUTHENTICATION = True
# SILKY_AUTHORISATION = True



# Generated by Django 1.9.7 on 2016-07-08 13:19

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Blind',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(upload_to=b'products')),
                ('name', models.TextField()),
                ('child_safe', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import os

from setuptools import setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-silk',
    use_scm_version=True,
    packages=['silk'],
    include_package_data=True,
    license='MIT License',
    description='Silky smooth profiling for the Django Framework',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/jazzband/django-silk',
    author='Michael Ford',
    author_email='mtford@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=2.2',
        'python-dateutil',
        'requests',
        'sqlparse',
        'Jinja2',
        'autopep8',
        'pytz',
        'gprof2dot>=2017.09.19',
    ],
    python_requires='>=3.7',
    setup_requires=['setuptools_scm'],
)



import django
from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("django-silk").version
except DistributionNotFound:
    pass

if django.VERSION < (3, 2):
    default_app_config = "silk.apps.SilkAppConfig"



from django.core.management.base import BaseCommand

import silk.models
from silk.utils.data_deletion import delete_model


class Command(BaseCommand):
    help = "Clears silk's log of requests."

    def handle(self, *args, **options):
        # Django takes a long time to traverse foreign key relations,
        # so delete in the order that makes it easy.
        delete_model(silk.models.Profile)
        delete_model(silk.models.SQLQuery)
        delete_model(silk.models.Response)
        delete_model(silk.models.Request)



import json
import os
import random
import traceback
from datetime import timedelta

from django.core import management
from django.utils import timezone

from silk import models
from silk.models import Profile, SQLQuery


class MockSuite:
    """
    Provides some fake data to play around with. Also useful for testing
    """
    methods = ['GET', 'POST', 'PUT', 'PATCH', 'HEAD', 'OPTIONS']
    path_components = ['path', 'to', 'somewhere', 'around', 'here', 'bobs', 'your', 'uncle']
    status_codes = [200, 201, 300, 301, 302, 403, 404, 500]
    profile_names = ['slow_bit_of_code', 'terrible_dependency', 'what_on_earth_is_this_code_doing']
    file_path = [os.path.realpath(__file__)]
    func_names = ['', '', '', 'foo', 'bar']
    view_names = ['app:blah', 'index', 'root', 'xxx:xyx']
    sql_queries = ['''
    SELECT Book.title AS Title,
    COUNT(*) AS Authors
     FROM  Book
     JOIN  Book_author
       ON  Book.isbn = Book_author.isbn
     GROUP BY Book.title;
 ''',
                   '''
    SELECT * FROM table
    ''', '''
    SELECT *
    FROM  Book
    WHERE price > 100.00
    ORDER BY title;
  ''', '''
  SELECT title,
       COUNT(*) AS Authors
 FROM  Book
 NATURAL JOIN Book_author
 GROUP BY title;
  ''',
                   '''
                   SELECT A.Col1, A.Col2, B.Col1,B.Col2
      FROM (SELECT RealTableZ.Col1, RealTableY.Col2, RealTableY.ID AS ID
              FROM RealTableZ
   LEFT OUTER JOIN RealTableY
                ON RealTableZ.ForeignKeyY=RealTableY.ID
             WHERE RealTableY.Col11>14
            ) AS B
        INNER JOIN A
                ON A.ForeignKeyY=B.ID
                   ''']

    response_content_types = ['text/html', 'application/json', 'text/css']
    response_content = {
        'text/html': ['<html></html>'],
        'text/css': ['#blah {font-weight: bold}'],
        'application/json': ['[1, 2, 3]']
    }
    request_content_types = ['application/json']
    request_content = {
        'application/json': ['{"blah": 5}']
    }

    def _random_method(self):
        return random.choice(self.methods)

    def _random_path(self):
        num_components = random.randint(1, 5)
        return '/' + '/'.join(random.sample(self.path_components, num_components)) + '/'

    def _random_query(self):
        return random.choice(self.sql_queries)

    def mock_sql_queries(self, request=None, profile=None, n=1, as_dict=False):
        start_time, end_time = self._random_time()
        queries = []
        for _ in range(0, n):
            tb = ''.join(reversed(traceback.format_stack()))
            d = {
                'query': self._random_query(),
                'start_time': start_time,
                'end_time': end_time,
                'request': request,
                'traceback': tb
            }
            if as_dict:
                queries.append(d)
            else:
                query = SQLQuery.objects.create(**d)
                queries.append(query)
        if profile:
            if as_dict:
                for q in queries:
                    profile['queries'].append(q)
            else:
                profile.queries.set(queries)
        return queries

    def mock_profile(self, request=None):
        start_time, end_time = self._random_time()
        dynamic = random.choice([True, False])
        profile = Profile.objects.create(start_time=start_time,
                                         end_time=end_time,
                                         request=request,
                                         name=random.choice(self.profile_names),
                                         file_path=random.choice(self.file_path),
                                         line_num=3,
                                         func_name=random.choice(self.func_names),
                                         dynamic=dynamic,
                                         end_line_num=6 if dynamic else None,
                                         exception_raised=random.choice([True, False])
                                         )
        self.mock_sql_queries(profile=profile, n=random.randint(0, 10))
        return profile

    def mock_profiles(self, request=None, n=1):
        profiles = []

        for _ in range(0, n):
            profile = self.mock_profile(request)
            profiles.append(profile)
        return profiles

    def _random_time(self):
        start_time = timezone.now()
        duration = timedelta(milliseconds=random.randint(0, 3000))
        end_time = start_time + duration
        return start_time, end_time

    def mock_request(self):
        start_time, end_time = self._random_time()
        num_sql_queries = random.randint(0, 20)
        request_content_type = random.choice(self.request_content_types)
        request_body = random.choice(self.request_content[request_content_type])
        time_taken = end_time - start_time
        time_taken = time_taken.total_seconds()
        request = models.Request.objects.create(method=self._random_method(),
                                                path=self._random_path(),
                                                num_sql_queries=num_sql_queries,
                                                start_time=start_time,
                                                end_time=end_time,
                                                view_name=random.choice(self.view_names),
                                                time_taken=time_taken,
                                                encoded_headers=json.dumps({'content-type': request_content_type}),
                                                body=request_body)
        response_content_type = random.choice(self.response_content_types)
        response_body = random.choice(self.response_content[response_content_type])
        models.Response.objects.create(request=request,
                                       status_code=random.choice(self.status_codes),
                                       encoded_headers=json.dumps({'content-type': response_content_type}),
                                       body=response_body)
        self.mock_sql_queries(request=request, n=num_sql_queries)
        self.mock_profiles(request, random.randint(0, 2))
        return request


if __name__ == '__main__':
    management.call_command('flush', interactive=False)
    requests = [MockSuite().mock_request() for _ in range(0, 100)]



from collections import namedtuple

from django.test import TestCase

from silk.views.code import _code, _code_context, _code_context_from_request

FILE_PATH = __file__
LINE_NUM = 5
END_LINE_NUM = 10

with open(__file__) as f:
    ACTUAL_LINES = [line + '\n' for line in f.read().split('\n')]


class CodeTestCase(TestCase):

    def assertActualLineEqual(self, actual_line, end_line_num=None):
        expected_actual_line = ACTUAL_LINES[LINE_NUM - 1:end_line_num or LINE_NUM]
        self.assertEqual(actual_line, expected_actual_line)

    def assertCodeEqual(self, code):
        expected_code = [line.strip('\n') for line in ACTUAL_LINES[0:LINE_NUM + 10]] + ['']
        self.assertEqual(code, expected_code)

    def test_code(self):
        for end_line_num in None, END_LINE_NUM:
            actual_line, code = _code(FILE_PATH, LINE_NUM, end_line_num)
            self.assertActualLineEqual(actual_line, end_line_num)
            self.assertCodeEqual(code)

    def test_code_context(self):
        for end_line_num in None, END_LINE_NUM:
            for prefix in '', 'salchicha_':
                context = _code_context(FILE_PATH, LINE_NUM, end_line_num, prefix)
                self.assertActualLineEqual(context[prefix + 'actual_line'], end_line_num)
                self.assertCodeEqual(context[prefix + 'code'])
                self.assertEqual(context[prefix + 'file_path'], FILE_PATH)
                self.assertEqual(context[prefix + 'line_num'], LINE_NUM)

    def test_code_context_from_request(self):
        for end_line_num in None, END_LINE_NUM:
            for prefix in '', 'salchicha_':
                request = namedtuple('Request', 'GET')(dict(file_path=FILE_PATH, line_num=LINE_NUM))
                context = _code_context_from_request(request, end_line_num, prefix)
                self.assertActualLineEqual(context[prefix + 'actual_line'], end_line_num)
                self.assertCodeEqual(context[prefix + 'code'])
                self.assertEqual(context[prefix + 'file_path'], FILE_PATH)
                self.assertEqual(context[prefix + 'line_num'], LINE_NUM)



"""WSGI config for django_silky project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/howto/deployment/wsgi/

"""
import os

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

from django.core.wsgi import get_wsgi_application  # noqa: E402

application = get_wsgi_application()



import logging

from django.http import HttpResponse
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request

Logger = logging.getLogger('silk.views.raw')


class Raw(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, request_id):
        typ = request.GET.get('typ', None)
        subtyp = request.GET.get('subtyp', None)
        body = None
        if typ and subtyp:
            silk_request = Request.objects.get(pk=request_id)
            if typ == 'request':
                body = silk_request.raw_body if subtyp == 'raw' else silk_request.body
            elif typ == 'response':
                Logger.debug(silk_request.response.raw_body_decoded)
                body = silk_request.response.raw_body_decoded if subtyp == 'raw' else silk_request.response.body
            return render(request, 'silk/raw.html', {
                'body': body
            })
        else:
            return HttpResponse(content='Bad Request', status=400)



from unittest.mock import NonCallableMock

from django.test import TestCase

from silk.collector import DataCollector
from silk.config import SilkyConfig
from silk.middleware import SilkyMiddleware
from silk.models import Request

from .util import delete_all_models


def fake_get_response():
    def fake_response():
        return 'hello world'
    return fake_response


class TestConfigMeta(TestCase):
    def _mock_response(self):
        response = NonCallableMock()
        response.headers = {}
        response.status_code = 200
        response.queries = []
        response.get = response.headers.get
        response.content = ''
        return response

    def _execute_request(self):
        delete_all_models(Request)
        DataCollector().configure(Request.objects.create())
        response = self._mock_response()
        SilkyMiddleware(fake_get_response)._process_response('', response)
        self.assertTrue(response.status_code == 200)
        objs = Request.objects.all()
        self.assertEqual(objs.count(), 1)
        r = objs[0]
        return r

    def test_enabled(self):
        SilkyConfig().SILKY_META = True
        r = self._execute_request()
        self.assertTrue(r.meta_time is not None
                        or r.meta_num_queries is not None
                        or r.meta_time_spent_queries is not None)

    def test_disabled(self):
        SilkyConfig().SILKY_META = False
        r = self._execute_request()
        self.assertFalse(r.meta_time)



# Generated by Django 2.0 on 2017-12-28 14:21

from django.db import migrations, models

import silk.storage


class Migration(migrations.Migration):

    dependencies = [
        ('silk', '0005_increase_request_prof_file_length'),
    ]

    operations = [
        migrations.AlterField(
            model_name='request',
            name='prof_file',
            field=models.FileField(blank=True, max_length=300, storage=silk.storage.ProfilerResultStorage(), upload_to=''),
        ),
    ]



from django.test import TestCase

from silk.collector import DataCollector


class TestCollector(TestCase):
    def test_singleton(self):
        a = DataCollector()
        b = DataCollector()
        c = DataCollector()
        self.assertTrue(a == b == c)

    def test_query_registration(self):
        mock_query = {}
        DataCollector().register_query(mock_query)
        self.assertIn(mock_query, list(DataCollector().queries.values()))

    def test_clear(self):
        self.test_query_registration()
        DataCollector().clear()
        self.assertFalse(DataCollector().queries)
